/*
 *		System News Displayer
 *
 *		(c) 1996 Martin Mares <mj@k332.feld.cvut.cz>
 *
 *		This program can be distributed under the terms of the GNU
 *		General Public License (see file COPYING). No warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#include <time.h>

static char path[PATH_MAX];
static time_t lasttime;

struct news_rec {
	struct news_rec *next;
	time_t time;
	uid_t uid;
	char filename[1];
	};

static struct news_rec *first;
static int news_count;

static void *
xalloc(int size)
{
void *p;
	p = malloc(size);
	if (!p)
		{
		fprintf(stderr,"news: Out of memory");
		exit(1);
		}
	return p;
}

static void
get_news_date(int restart)
{
char *h = getenv("HOME");
struct stat st;
	if (h)
		{
		strcpy(path, h);
		strcat(path, "/");
		}
	strcat(path, ".newsdate");
	if (!restart && !stat(path, &st) && S_ISREG(st.st_mode))
		lasttime = st.st_mtime;
}

static void
scan_news(void)
{
DIR *d;
struct dirent *e;
struct news_rec *n;
struct stat st;
	if (!chdir("/var/spool/news") && (d = opendir(".")))
		{
		while (e = readdir(d))
			{
			if (!stat(e->d_name, &st) && S_ISREG(st.st_mode))
				{
				time_t age = st.st_mtime;
				if (age >= lasttime)
					{
					n = xalloc(sizeof(struct news_rec) + strlen(e->d_name));
					n->next = first;
					first = n;
					n->time = age;
					strcpy(n->filename, e->d_name);
					n->uid = st.st_uid;
					news_count++;
					}
				}
			}
		closedir(d);
		}
	else
		{
		perror("Cannot open /var/spool/news");
		exit(1);
		}
}

static int
compar(const struct news_rec **a, const struct news_rec **b)
{
int aa = (*a)->time;
int bb = (*b)->time;
	if (aa < bb)
		return -1;
	else if (aa > bb)
		return 1;
	else
		return 0;
}

static void
sort_news(void)
{
struct news_rec **index, **last;
int i;
	index = xalloc(sizeof(struct news_rec *) * news_count);
	for (i=0; i<news_count; i++)
		{
		index[i] = first;
		first = first->next;
		}
	qsort(index, news_count, sizeof(struct news_rec *),
		(__compar_fn_t) compar);
	last = &first;
	for (i=0; i<news_count; i++)
		{
		*last = index[i];
		last = &index[i]->next;
		}
	*last = NULL;
}

static void
print_news(void)
{
int fd[2];
int pid, stat, FD;
char *buff, *t, *s;
static struct passwd *p;
	if (pipe(fd))
		{
		perror("Cannot create pipe");
		exit(1);
		}
	pid = fork();
	if (pid == -1)
		{
		perror("Fork failed");
		exit(1);
		}
	if (pid)
		{
		close(fd[0]);
		buff = xalloc(4096);
		while (first)
			{
			p = getpwuid(first->uid);
			t = ctime(&first->time);
			if (s = strchr(t, '\n'))
				*s = 0;
			sprintf(buff, "From %s (%s): ", (p ? p->pw_name : "unknown user"), t);
			if (write(fd[1], buff, strlen(buff)) < 0)
				break;
			FD = open(first->filename, O_RDONLY);
			if (FD == -1)
				{
				sprintf(buff, "Unable to open %s\n", first->filename);
				write(fd[1], buff, strlen(buff));
				}
			else
				{
				while ((stat = read(FD, buff, 4096)) > 0)
					write(fd[1], buff, stat);
				close(FD);
				}
			if (write(fd[1], "\n", 1) < 0)
				break;
			first = first->next;
			}
		}
	else
		{
		char *argv[2];
		close(0); dup(fd[0]); close(fd[1]); close(fd[0]);
		argv[0] = "less";
		argv[1] = NULL;
		execvp(argv[0], argv);
		exit(1);
		}
	close(fd[1]);
	wait(&stat);
}

static void
print_short_news(void)
{
FILE *f;
char linebuf[256];
char *c;
	if (first)
		{
		while (first)
			{
			f = fopen(first->filename, "r");
			if (f)
				{
				fgets(linebuf, 256, f);
				if (c = strchr(linebuf, '\n'))
					*c = 0;
				printf("News: %s\n", linebuf);
				fclose(f);
				}
			else
				printf("%s: unable to open (%s)\n", first->filename, strerror(errno));
			first=first->next;
			}
		putchar('\n');
		}
}

static void
save_news_date(void)
{
int fd;
	fd = open(path, O_CREAT | O_RDWR, 0644);
	if (fd < 0)
		{
		fprintf(stderr, "Cannot create %s: %s\n", path, strerror(errno));
		exit(1);
		}
	ftruncate(fd, 0);
	close(fd);
}

void
main(int argc, char **argv)
{
int mode, restart;
	mode = 0;
	restart = 0;
	while (argc > 1)
		{
		char *a = argv[argc-1];
		if (!strcmp(a, "-i"))
			mode = 1;
		else if (!strcmp(a, "-r"))
			restart = 1;
		else
			{
			fprintf(stderr, "Usage: news [-i] [-r]\n");
			exit(1);
			}
		argc--;
		}
	get_news_date(restart);
	scan_news();
	sort_news();
	if (mode)
		print_short_news();
	else
		{
		if (first)
			print_news();
		}
	if (!mode || restart)
		save_news_date();
}
